%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This work is licensed under a Creative Commons Attribution 4.0
% International License. http://creativecommons.org/licenses/by/4.0/
% The authors do not assume liability for errors contained in or for
% damages arising from the use of the software.
%
% CC-BY (4.0) Oliver Taubmann (oliver.taubmann@fau.de), 2015
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This script facilitates batch mode comparison of several image pairs.

%% Settings

settings.prof_filter_size = 5;   % Size of filter kernel for smoothing profiles
settings.es_filter_size   = 20;  % Size of filter kernel for smoothing ES values
settings.prof_spacing     = 0.5; % Spacing of profiles in voxels
settings.prof_len         = 5;   % Length of profiles in voxels
settings.num_samples      = settings.prof_len * 4; % Number of samples per profile

%% Runs

output = 'output'; % Output filename (w/o extension)

runs = cell(0);

run.label        = 'Run 1';
run.segmentation = 'segmentation_run_1.mat';
run.ref_image    = 'Path/To/Run1/ReferenceImage';
run.cmp_image    = 'Path/To/Run1/ComparedImage';
runs = [runs, run];

run.label        = 'Run 2';
run.segmentation = 'segmentation_run_2.mat';
run.ref_image    = 'Path/To/Run2/ReferenceImage';
run.cmp_image    = 'Path/To/Run2/ComparedImage';
runs = [runs, run];

run.label        = 'Run 3';
run.segmentation = 'segmentation_run_3.mat';
run.ref_image    = 'Path/To/Run3/ReferenceImage';
run.cmp_image    = 'Path/To/Run3/ComparedImage';
runs = [runs, run];

% You can add more runs here

for run = 1:length(runs)
    
    %% Load images and segmentation
    
    % (May be replaced by other file format loaders as needed)
    volume1 = read_mhd(runs{run}.ref_image);
    volume2 = read_mhd(runs{run}.cmp_image);
        
    % Create segmentation if it doesn't exist
    if ~exist(runs{run}.segmentation, 'file')     
        
        disp(['Segmentation file for run "' runs{run}.label ...
              '" not found, starting manual segmentation']);
        
        z_index = [];
        
        while isempty(z_index)
            answer = inputdlg('Which z-slice do you want to segment? (1-based index)', 'Segmentation');
            z_index = round(str2num(answer{1}));
            
            if isempty(z_index)
                fprintf(2, 'Index is not numeric, try again!\n');
                continue;
            end
            if ~(z_index >= 1 && z_index <= volume1.size(3))
                fprintf(2, 'Index is out of valid range [%d..%d], try again!\n', 1, volume1.size(3));
                z_index = [];
            end
        end
        
        slice1 = double(volume1.data(:, :, z_index)');
        slice2 = double(volume2.data(:, :, z_index)');
        slices = {slice1, slice2};
        
        figure(1);
        hold off;

        for i = 1:2
            subplot(1, 2, i);
            imagesc(slices{i});
            colormap(gray);
            axis('image');
            hold on;
            title(sprintf('Slice %d', i));
        end
        
        [xs, ys] = segment_edge();
        
        close;
        
        save(runs{run}.segmentation, 'xs', 'ys', 'z_index');
    end
    
    load(runs{run}.segmentation, 'xs', 'ys', 'z_index');
    
    slice1 = double(volume1.data(:, :, z_index)');
    slice2 = double(volume2.data(:, :, z_index)');
    
    slices = {slice1, slice2};
        
    %% Create profiles for segmentation
    
    [prof_ps, ps] = create_profiles(xs, ys, settings.prof_len, settings.prof_spacing);
        
    %% Sample profiles
    
    profs = cell(1, 2);
    
    for i = 1:2
        profs{i} = sample_profiles(prof_ps, slices{i}, settings.num_samples);
        profs{i} = medfilt2(profs{i}, [1 settings.prof_filter_size], 'symmetric'); % Remove outliers
    end
    
    %% Compute edge sharpness on profiles
    
    ess       = cell(1, 2);
    ess_filt  = cell(1, 2);
    median_es = cell(1, 2);
    
    for i = 1:2
        ess{i}       = compute_es(profs{i}, settings.prof_len);
        ess_filt{i}  = medfilt2(ess{i}, [settings.es_filter_size, 1], 'symmetric');
        median_es{i} = median(ess{i});
    end

    runs{run}.median_es_ref = median_es{1};
    runs{run}.median_es_cmp = median_es{2};
    runs{run}.esi           = median_es{2} / median_es{1};
end

%% Write results to file

fileID = fopen([output '.txt'], 'w');

save([output '.mat'], 'settings', 'runs');

runs = [{settings} runs]; % add settings struct for printing

fprintf(fileID, 'Settings:\n');

for run = 1:length(runs)  
    
    fields = fieldnames(runs{run});
    
	for i = 1:numel(fields)
        field = fields{i};
        val   = num2str(runs{run}.(field));
        val(ismember(val, '\\')) = '/';
        fprintf(fileID, [field ': ']);
        fprintf(fileID, [val '\n']);
	end
    
    fprintf(fileID, '\n');
end

fclose(fileID);